package com.sillycycle.bagleyd.mball;

/*
 * @(#)MballSolve.java
 *
 * Many ideas found at
 *  http://wsd.com/masterball now offline, but thanks to wayback machine see
 *  http://web.archive.org/web/20041025133454/wsd.com/masterball
 * Also
 *  http://www.permutationpuzzles.org/rubik/mball/rainbow.html
 *  just make sure you subtract 1 as I start from 0.
 * And
 *  http://www.jaapsch.net/puzzles/masterball.html
 *
 * Copyright 2010 - 2018  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>PyraminxSolve</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	  David A. Bagley
 * @author	  bagleyd AT verizon.net
 * @author	  https://www.sillycycle.com/puzzles.html
 */

class MballSolve {
	private static JFrame frame;
	private static MballCanvas mball;
	private static final boolean DEBUG = false;
	private static Random generator;
	//private static int counter; // debugging

	MballSolve(JFrame frame, MballCanvas mball) {
		MballSolve.frame = frame;
		MballSolve.mball = mball;
		generator = new Random(System.nanoTime());
		//this.counter = 0;
	}

	boolean moveAPiece(int wedge, int band, int direction, boolean control) {
		if (!getFlag()) {
			return false;
		}
		mball.movePuzzleDelay(wedge, band, direction,
			control, MballCanvas.DOUBLE);
		return true;
	}

	/* r-> rotate */
	void rCW(int b) {
		moveAPiece(0, b, MballCanvas.CW, false);
	}

	void rCCW(int b) {
		moveAPiece(0, b, MballCanvas.CCW, false);
	}

	/* f -> flip */
	void f(int i) {
		moveAPiece(i, 0, MballCanvas.mapWedgeToDir[(mball.wedges -
			MballInterface.MIN_WEDGES) / 2][i], false);
	}

	void fo(int i, int o) {
		f((i + o + mball.wedges) % mball.wedges);
	}

	void rPuzzleCW(int b) {
		moveAPiece(0, b, MballCanvas.CW, true);
	}

	void rPuzzleCCW(int b) {
		moveAPiece(0, b, MballCanvas.CCW, true);
	}

	void fPuzzle(int i) {
		moveAPiece(i, 0, MballCanvas.mapWedgeToDir[(mball.wedges -
			MballInterface.MIN_WEDGES) / 2][i], true);
	}

	static int ringAdd(int i, int j) {
		return (i + j + mball.wedges) % mball.wedges;
	}

	void rnCW(int b, int n) {
		int i;

		for (i = 0; i < n; i++) {
			rCW(b);
		}
	}

	void rnCCW(int b, int n) {
		int i;

		for (i = 0; i < n; i++) {
			rCCW(b);
		}
	}

	void rnHalf(int b) {
		int i;

		if (generator.nextInt(2) == 0) {
			for (i = 0; i < mball.wedges / 2; i++) {
				rCW(b);
			}
		} else {
			for (i = 0; i < mball.wedges / 2; i++) {
				rCCW(b);
			}
		}
	}

	void rNHalfCW() {
		int i;

		for (i = 0; i < mball.bands / 2; i++) {
			rCW(i);
		}
	}

	void rNHalfCCW() {
		int i;

		for (i = 0; i < mball.bands / 2; i++) {
			rCCW(i);
		}
	}

	void rNHalfHalf() {
		if (mball.generator.nextInt(2) == 0) {
			rNHalfCW();
		} else {
			rNHalfCCW();
		}
	}

	void rSHalfCW() {
		int i;

		for (i = mball.bands - 1; i >= (mball.bands + 1) / 2; i--) {
			rCW(i);
		}
	}

	void rSHalfCCW() {
		int i;

		for (i = mball.bands - 1; i >= (mball.bands + 1) / 2; i--) {
			rCCW(i);
		}
	}

	void rSHalfHalf() {
		if (mball.generator.nextInt(2) == 0) {
			rSHalfCW();
		} else {
			rSHalfCCW();
		}
	}

	void rNPoleCW(int b) {
		int i;

		for (i = 0; i <= b; i++) {
			rCW(i);
		}
	}

	void rNPoleCCW(int b) {
		int i;

		for (i = 0; i <= b; i++) {
			rCCW(i);
		}
	}

	static boolean allTop() {
		int tz, band;

		for (tz = 0; tz < mball.wedges; tz++) {
			for (band = 0; band < mball.bands; band++) {
				if (mball.wedgeLoc[tz + mball.wedges + band].direction == 1) {
					return false;
				}
			}
		}
		return true;
	}

	static int wedgeFromPosition(int position) {
		return position % mball.wedges;
	}

	static int bandFromPosition(int position) {
		return position / mball.wedges;
	}

	static int findColor(int wedgeColor, int band) {
		int i, tz, b;

		for (i = 0; i < 2; i++) {
			/* only check the 2 mball.bands that are possible */
			if (i == 0)
				b = band;
			else
				b = mball.bands - 1 - band;
			for (tz = 0; tz < mball.wedges; tz++) {
				if (wedgeColor == mball.wedgeLoc[tz + mball.wedges * b].wedge) {
					if (DEBUG)
						System.out.println("Color for wedge " +
							tz + ", band " + b + " is " + wedgeColor);
					if ((mball.orient) &&
							((i == 0 && mball.wedgeLoc[tz + mball.wedges * b].direction == 1)
								|| (i == 1 && mball.wedgeLoc[tz + mball.wedges * b].direction == 0)
								|| (tz <= wedgeColor && b < band))) /* finding solved portion */
						continue;
					return tz + b * mball.wedges;
				}
			}
		}
		System.out.println("Color not found for " + wedgeColor + " " + band);
		return mball.wedges * mball.bands;
	}

	static int findOppSameLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges / 2; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[tz + mball.wedges / 2 + mball.wedges * band].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static int findAdjOppLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[(tz + 1) % mball.wedges + mball.wedges * (mball.bands - 1 - band)].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static int findNextAdjSameLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[(tz + 2) % mball.wedges + mball.wedges * band].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static int findAdjOppOppLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[(tz + mball.wedges / 2 - 1) % mball.wedges + mball.wedges * (mball.bands - 1 - band)].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static int findAdjSameLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[(tz + 1) % mball.wedges + mball.wedges * band].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static int findOppOppLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[(tz + mball.wedges / 2) % mball.wedges + mball.wedges * (mball.bands - 1 - band)].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static int findAdjOppSameLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[(tz + mball.wedges / 2 - 1) % mball.wedges + mball.wedges * band].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static int findNextAdjOppLat(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge ==
					mball.wedgeLoc[(tz + 2) % mball.wedges + mball.wedges * (mball.bands - 1 - band)].wedge) {
				return tz;
			}
		}
		return mball.wedges;
	}

	static boolean checkBandSolved(int band) {
		int tz;

		for (tz = 0; tz < mball.wedges - 1; tz++) {
			if (mball.wedgeLoc[tz + mball.wedges * band].wedge !=
					mball.wedgeLoc[tz + mball.wedges * (mball.bands - 1 - band)].wedge) {
				return false;
			}
		}
		return true;
	}

	static boolean checkSegmentSolved(int wedge, int band, int lat) {
		return (mball.wedgeLoc[wedge + mball.wedges * lat].wedge ==
				mball.wedgeLoc[wedge + mball.wedges * band].wedge);
	}

	/* swaps 4 for 4 wedges, diagonal */
	void swap4by4d() {
		rCCW(0);
		f(0);
		rCCW(0);
		f(0);
		rCW(0);
		f(3);
		rCW(0);
		f(3);
		rCCW(0);
		f(0);
		rCCW(0);
		f(0);
		rCW(0);
		f(3);
		rCW(0);
		f(3);
		rCCW(0);
		f(0);
		rCCW(0);
		f(0);
		rCW(0);
		f(3);
		rCW(0);
		f(3);
		rCCW(0);
		f(0);
		rCCW(0);
		f(0);
	}

	/* swaps 4 for 4 wedges, adj, moves 2 to bottom and 2 to top */
	void swap4by4() {
		rCW(0);
		f(0);
		rnCW(0, 2);
		f(0);
		rnCW(0, 2);
		f(0);
		rCW(0);
		f(0);
	}

	/* swaps 2 for 4 wedges, swap 2&3, 30 moves */
	/* Interesting thing is in a different order (say cw first then ccw) */
	/* it is period 24 but here its period 60. */
	void swap2by4() {
		for (int i = 0; i < 7; i++) {
			rCCW(0);
			f(2);
			rCW(0);
			f(3);
		}
		rCCW(0);
		f(2);
	}

	/* 1,0 -> 2,0  1,1 -> 1,0  2, 0 -> 1, 1 */
	void flipFor4_3(int wedge, int band) {
		/* 0 1 -1  0 -1 0  0 -1 0  0 1 1  0 -1 0  0 -1 0 */
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCW(band);
	}

	/* flip bottom 0 & 1 wedge */
	void flipFor4a(int wedge, int band) {
		if (DEBUG)
			System.out.println("flipFor4a " +
				wedge + " " + band);
		/* 0 0 -1  0 -1 0  0 1 0  0 1 1  0 1 0  0 -1 0 */
		f(wedge);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCW(band);
	}

	/* flip first for 4, flip 0th top wedge and 3rd bottom (diagonal) */
	void flipFor4d(int wedge, int band) {
		/* 0 1 0  0 1 -1  0 -1 0  0 1 -1  0 1 0  0 1 -1 */
		if (DEBUG)
			System.out.println("flipFor4d " +
				wedge + " " + band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
	}

	/* flip 0th wedge */
	void flipFor4w(int wedge, int band) {
		/* 0 0 -1  0 1 0  0 1 -1  0 1 0  0 1 -1  0 -1 0  0 1 0  0 0 0 */
		if (DEBUG)
			System.out.println("flipFor4w " +
				wedge + " " + band);
		f(wedge);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCCW(band);
		f(wedge);
	}

	/* flip 4 0th & 5th wedges */
	void flipFor6_4(int wedge, int band) {
		/* 0 0 -1  0 -1 1  0 1 0  0 0 0 */
		if (DEBUG)
			System.out.println("flipFor6_4 " +
				wedge + " " + band);
		f(wedge);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		f(wedge);
	}

	/* flip 2 3rd bottom & 5th top wedges */
	void flipFor6_4s(int wedge, int band) {
		/* 0 1 2  0 1 0  0 -1 -2  0 1 2  0 2 0  0 -1 -2 */
		if (DEBUG)
			System.out.println("flipFor6_4s " +
				wedge + " " + band);
		f(wedge);
		rCCW(band);
		rnCCW(mball.bands - 1 - band, 2);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCW(band);
		rnCW(mball.bands - 1 - band, 2);
		f(wedge);
		rCCW(band);
		rnCCW(mball.bands - 1 - band, 2);
		f(wedge);
		rnCCW(band, 2);
		f(wedge);
		rCW(band);
		rnCW(mball.bands - 1 - band, 2);
	}

	/* flip 4 1st & 5th wedges */
	void flipFor6_4sm(int wedge, int band) {
		/* 0 0 0  0 1 0  0 -2 0  0 1 -1  0 1 1  0 -1 0 */
		if (DEBUG)
			System.out.println("flipFor6_4sm " +
				wedge + " " + band);
		rCCW(band);
		f(wedge);
		rnCW(band, 2);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
	}

	/* flip 1st top & 1st bottom */
	void flipFor6w(int wedge, int band) {
		if (DEBUG)
			System.out.println("flipFor6w " +
				wedge + " " + band);
		/* 0 1 2  0 -1 -1  0 1 2  0 -1 -2  0 1 -2  0 -1 -2 */
		f(wedge);
		rCCW(band);
		rnCCW(mball.bands - 1 - band, 2);
		f(wedge);
		rCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		rnCCW(mball.bands - 1 - band, 2);
		f(wedge);
		rCW(band);
		rnCW(mball.bands - 1 - band, 2);
		f(wedge);
		rCCW(band);
		rnCW(mball.bands - 1 - band, 2);
		f(wedge);
		rCW(band);
		rnCW(mball.bands - 1 - band, 2);
	}

	/* 123 -> 231 */
	void flipFor6_3(int wedge, int band) {
		/* 0 1 -1  0 1 0  0 -1 0  0 1 0  0 1 0  0 0 0  0 -1 1  0 -1 -1  0 -1 1  0 0 0 */
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
	}

	/* 123 -> 312 */
	void flipFor6_3r(int wedge, int band) {
		f(wedge);
		rCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
	}

	/* flip 3 4th top & bottom & 5th wedge top */
	void flipFor6_3f(int wedge, int band) {
		/* 0 0 1  0 -1 1  0 1 -1  0 -1 0  0 -1 0  0 1 0  0 1 -1  0 0 0 */
		f(wedge);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCW(band);
		f(wedge);
		rCCW(band);
		f(wedge);
		rCCW(band);
		rCW(mball.bands - 1 - band);
		f(wedge);
	}

	/* flip 0th top & 5th bottom */
	void flipFor6d(int wedge, int band) {
		if (DEBUG)
			System.out.println("flipFor6d " +
				wedge + " " + band);
		/* 0 0 1  0 -1 -2  0 -2 -1  0 -2 0  0 2 1  0 1 2 */
		f(wedge);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCW(band);
		rnCW(mball.bands - 1 - band, 2);
		f(wedge);
		rnCW(band, 2);
		rCW(mball.bands - 1 - band);
		f(wedge);
		rnCW(band, 2);
		f(wedge);
		rnCCW(band, 2);
		rCCW(mball.bands - 1 - band);
		f(wedge);
		rCCW(band);
		rnCCW(mball.bands - 1 - band, 2);
	}

/* I found this 8+4 move one... ok probably not the first.  :)
   it flips 3, 4 both on top and flips 7 on top with 3 on bottom */
	void simpleSwap() {
		rnCW(2, 4);
		rnCW(3, 4);
		rCW(0);
		f(0);
		rCCW(0);
		f(0);
		rCCW(0);
		f(0);
		rCW(0);
		f(0);
		rnCW(3, 4);
		rnCW(2, 4);
	}

/* post part of threeOut move I found */
	void threeOut1(int wedge, int band) {
		int b = mball.bands - 1 - band;
		int tz = (wedge + 7) % mball.wedges;
		int lon = wedge;

/* f2 *r1*f1 *r1^(-1)*f1 *r4^(-1)*f1 *r4*f1 *f2 */
		rCW(band);f(lon);
		rCCW(band); f(lon);
		rCCW(b); f(lon);
		rCW(b); f(lon);
		f(tz);
	}

/* I found this 10 move for 3 out by accident.  :)
   2 moves added to make it pretty. */
/* lower band 2 -> 4, 3 -> 2, 4 -> 3; 10 moves */
	void threeOut(int wedge, int band) {
		int tz = (wedge + 7) % mball.wedges;

/* f2 *r1*f1 *r1^(-1)*f1 *r4^(-1)*f1 *r4*f1 *f2 */
		f(tz);
		threeOut1(wedge, band);
	}

/* Combining above threeOut and a rotation. */
/* lower band 2 -> 5, 5 -> 2; 31 moves */
	void twoOut(int wedge, int band) {
/* f2 *r1*f1 *r1^(-1)*f1 *r4^(-1)*f1 *r4*f1 *f2
*f7 *r1*f6 *r1^(-1)*f6 *r4^(-1)*f6 *r4*f6 *f7
*f5 *r1*f4 *r1^(-1)*f4 *r4^(-1)*f4 *r4*f4 *f5
*r4 */
		threeOut(wedge, band);
		threeOut((wedge + 3) % mball.wedges, band);
		threeOut((wedge + 5) % mball.wedges, band);
		rCW(mball.bands - 1 - band);
	}

/* Small changes to help save 2 moves */
	void twoOut1(int wedge, int band) {
/* r1*f1 *r1^(-1)*f1 *r4^(-1)*f1 *r4*f1 *f2
*f7 *r1*f6 *r1^(-1)*f6 *r4^(-1)*f6 *r4*f6 *f7
*f5 *r1*f4 *r1^(-1)*f4 *r4^(-1)*f4 *r4*f4 *f5
*r4 */
		threeOut1((wedge + 1) % mball.wedges, band);
		threeOut((wedge + 4) % mball.wedges, band);
		threeOut((wedge + 6) % mball.wedges, band);
		rCW(mball.bands - 1 - band);
	}

/* 37 moves to flip just one wedge band */
	void flipWedgeSegment(int wedge, int band) {
		int tz = (wedge + 7) % mball.wedges;
		int lon = (wedge + 4) % mball.wedges;

/* r4^(-2)*f3
*f2 *r1*f1 *r1^(-1)*f1 *r4^(-1)*f1 *r4*f1 *f2
*f7 *r1*f6 *r1^(-1)*f6 *r4^(-1)*f6 *r4*f6 *f7
*f5 *r1*f4 *r1^(-1)*f4 *r4^(-1)*f4 *r4*f4 *f5
*r4
*f3 *r4^2 */
		rnCCW(band, 2); f(tz);
		twoOut(lon, band);
		f(tz);
		rnCW(band, 2);
	}

/* Slight improvement, 35 moves to flip just one wedge band */
	void flipWedgeSegment1(int wedge, int band) {
		int b = mball.bands - 1 - band;

/* r4^(-2)
	*r1*f0 *r1^(-1)*f0 *r4^(-1)*f0 *r4*f0 *f1
*f6 *r1*f5 *r1^(-1)*f5 *r4^(-1)*f5 *r4*f5 *f6
*f4 *r1*f3 *r1^(-1)*f3 *r4^(-1)*f3 *r4*f3 *f4
*r4
*f1 *r4^2 */
	rnCCW(b, 2);
	twoOut1(wedge, band);
	f(wedge);
	rnCW(b, 2);
}

/* saturn move got from www.jaapsch.net/puzzles/saturn.htm
  A/A/C/A/A A A/A/C/A/A
  Not sure if it works as described... seem to make then numbers
  counterclockwise and there is an extra twist.
*/
	void saturnMove(int wedge, int band) {
		int tz = (wedge + mball.wedges / 2) % mball.wedges;

		rCCW(band); f(tz);
		rCCW(band); f(tz);
		rCW(band); f(tz);
		rCCW(band); f(tz);
		rnCCW(band, 3); f(tz);
		rCCW(band); f(tz);
		rCW(band); f(tz);
		rCCW(band); f(tz);
		rCCW(band); f(tz);
}

/* Described by David Joyner works similar for both 10 and 12 wedges */
/* http://www.usna.edu/Users/math/wdj/mball/puck.htm */
	void puckMove(int wedge) {
/* (f1*f2*f3*f4*f5*f6)^2 */
		f(wedge);
		fo(wedge, 1);
		fo(wedge, 2);
		fo(wedge, 3);
		fo(wedge, 4);
		fo(wedge, 5);
		f(wedge);
		fo(wedge, 1);
		fo(wedge, 2);
		fo(wedge, 3);
		fo(wedge, 4);
		fo(wedge, 5);
	}

/* these Saturn moves are my own... I removed the rotation by changing
flips and took some starting moves above but changed to make work. */

/* Also 2 pair move is sometimes needed and this is not mentioned in
above web site. */
/* Saturn move 01/34 */
	void saturn2Pair(int tz) {
		fo(tz, 6);
		fo(tz, 5);
		fo(tz, 6);
		fo(tz, 5);
		fo(tz, 7);
		f(tz);
		fo(tz, 7);
		fo(tz, 6);
		fo(tz, 7);
		fo(tz, 5);
		f(tz);
		fo(tz, 6);
		fo(tz, 7);
	}

/* 0 -> 2, 12 -> 01_ */
	void saturn3Jump(int tz) {
		fo(tz, 6);
		fo(tz, 5);
		fo(tz, 6);
		fo(tz, 5);
		fo(tz, 7);
		fo(tz, 5);
		f(tz);
		fo(tz, 6);
		fo(tz, 5);
		fo(tz, 7);
		f(tz);
		fo(tz, 6);
		fo(tz, 7);
		f(tz);
		fo(tz, 7);
	}

/* 2 -> 0, 10_ -> 21 */
	void saturn3JumpBack(int tz) {
		fo(tz, 7);
		f(tz);
		fo(tz, 7);
		fo(tz, 6);
		f(tz);
		fo(tz, 7);
		fo(tz, 5);
		fo(tz, 6);
		f(tz);
		fo(tz, 5);
		fo(tz, 7);
		fo(tz, 5);
		fo(tz, 6);
		fo(tz, 5);
		fo(tz, 6);
	}

/* This will swap 2 wedges side by side, but only if even number of bands */
	void swapEven(int tz) {
		fo(tz, 5);
		rNHalfCCW();
		fo(tz, 5);
		rNHalfCW();
		rSHalfCCW();
		fo(tz, 5);
		rSHalfCW();
		fo(tz, 5);
	}

	void longSwapAdj34() {
/* f1 *r3*r4*f2*f4 *r1*r4^(-1)*f4 *r4^4*f4
   *r4*r1^(-1)*f4 *r4^4*f2 *r3^(-1)*r4^(-1)*f1 */
		f(0);
		rCW(2); rCW(3); f(7); f(5);
		rCW(0); rCCW(3); f(5);
		rCW(3); rCW(3); rCW(3); rCW(3); f(5);
		rCW(3); rCCW(0); f(5);
		rCW(3); rCW(3); rCW(3); rCW(3); f(7);
		rCCW(2); rCCW(3); f(0);
	}

/* swap wedges 3 & 4 for both bands */
	void swapAdj34() {
/* f1 *r1*f4 *r1^(-1)*r4*f4 *r4^(-1)*f1 */
		f(0);
		rCW(0); f(5);
		rCCW(0); rCW(3); f(5);
		rCCW(3); f(0);
	}

/* swap wedges from opposite sides for 0 & 4 for both bands */
	void swapOp04(int wedge, int band) {
/* r1*f4 *r1^(-1)*r4*f4 *r4^(-1) */
		int b = mball.bands - 1 - band;
		int tz = (wedge + 5) % mball.wedges;

		rCW(band); f(tz);
		rCCW(band); rCW(b); f(tz);
		rCCW(b);
	}

/* swapOp04 with rotation */
	void flipOppSegments(int wedge, int band) {
		rnHalf(band);
		swapOp04(wedge, band);
		rnHalf(band);
	}

/* Pretty Patterns */
	void quadrantizedPattern() {
/* r3^4*r4^4*f1 *r3^2*r4^2*f2 *r3^2*r4^2*f3 */
		rCW(2); rCW(2); rCW(2); rCW(2);
		rCW(3); rCW(3); rCW(3); rCW(3);
		f(0);
		rCW(2); rCW(2); rCW(3); rCW(3); f(7);
		rCW(2); rCW(2); rCW(3); rCW(3); f(6);
	}

	void bandedPattern() {
/* r1^4*r3^4*f1 *r1^2*r3^2*f2 *r1^2*r3^2*f3 */
		rCW(0); rCW(0); rCW(0); rCW(0);
		rCW(2); rCW(2); rCW(2); rCW(2);
		f(0);
		rCW(0); rCW(0); rCW(2); rCW(2); f(7);
		rCW(0); rCW(0); rCW(2); rCW(2); f(6);
	}

	void move0() {
/* move1=r1*r3*f1 *r1^(-1)*r3^(-1)*f7*f2, */
		rCW(0); rCW(2); f(0);
		rCCW(0); rCCW(2); f(2); f(7);
	}

	void move1() {
/* move2=r1*r3*f1 *r1^(-1)*r3^(-1)*f1, */
		rCW(0); rCW(2); f(0);
		rCCW(0); rCCW(2); f(0);
	}

	void move2() {
/* move3=r1*r3*f3 *r1^(-1)*r3^(-1)*f4*f5, */
		rCW(0); rCW(2); f(6);
		rCCW(0); rCCW(2); f(5); f(4);
	}

	void move3() {
/* move4=r1*r3*f2 *r1^(-1)*r3^(-1)*f2, */
		rCW(0); rCW(2); f(7);
		rCCW(0); rCCW(2); f(7);
	}

	void move4() {
/* move5=f3*f4*f5*f6 *f3*f4*f5*f6 *r1*r2*r3*r4 (swaps columns 1,2), */
		f(6); f(5); f(4); f(3);
		f(6); f(5); f(4); f(3);
		rCW(0); rCW(1); rCW(2); rCW(3);
	}

	void move5() {
/* move6=f4*f5*f6*f7 *f4*f5*f6*f7 *r1*r2*r3*r4 (swaps columns 2,3) */
		f(5); f(4); f(3); f(2);
		f(5); f(4); f(3); f(2);
		rCW(0); rCW(1); rCW(2); rCW(3);
	}

/* 50 moves */
	void checkeredPattern() {
/* checkered=move1*move2*move3*move4*move5*move6 */
		move0(); move1(); move2(); move3(); move4(); move5();
	}

/* 50 moves (5,6,7), not found in rainbow.doc but in rainbo12.log */
	void long3Cycle() {
/*long3Cycle=f2 *r1*f4 *r1^(-2)*f4 *r4^(-1)*f4
  *r1*f4 *r4^(-1)*f4 *r1^(-2)*f4 *r4*f4
  *r1^2*f4 *r4^(-1)*f4 *r1*f4 *r4^2*f4
  *r1^(-1)*f4 *r4^(-2)*f4 *r1^(-1)*f4 *r4*f4
  *r1^2*f4 *r4^(-1)*f4 *r1^(-1)*f4 *r4*f4
  *r1*f4 *r4*r1^(-1)*f2 */
		f(7);
		rCW(0); f(5);
		rCCW(0); rCCW(0); f(5);
		rCCW(3); f(5);
		rCW(0); f(5);
		rCCW(3); f(5);
		rCCW(0); rCCW(0); f(5);
		rCW(3); f(5);
		rCW(0); rCW(0); f(5);
		rCCW(3); f(5);
		rCW(0); f(5);
		rCW(3); rCW(3); f(5);
		rCCW(0); f(5);
		rCCW(3); rCCW(3); f(5);
		rCCW(0); f(5);
		rCW(3); f(5);
		rCW(0); rCW(0); f(5);
		rCCW(3); f(5);
		rCCW(0); f(5);
		rCW(3); f(5);
		rCW(0); f(5);
		rCW(3); rCCW(0); f(7);
	}

/* 7 * 44 moves -> 3 out */
	void long30utSwap() {
		f(7);
		rCW(0); f(5);
		rCCW(0); f(5);
		rCCW(3); f(5);
		rCW(0); f(5);
		rCCW(3); f(5);
		rCCW(0); f(5);
		rCW(3); f(5);
		rCCW(0); f(5);
		rCCW(3); f(5);
		rCW(0); f(5);
		rCCW(3); f(5);
		rCCW(0); f(5);
		rCCW(3); f(5);
		rCCW(0); f(5);
		rCW(3); f(5);
		rCCW(0); f(5);
		rCCW(3); f(5);
		rCCW(0); f(5);
		rCW(3); f(5);
		rCW(0); f(5);
		rCW(3); rCCW(0); f(7);
	}

	void polar2Swap16() {
/* polar2swap36=f1 *r3^(-1)*r4^(-1)*f1*f2 *r1*r4^(-1)*f2 *r4^4*f2
  *r1^(-1)*r4*f2 *r4^4*f1 *r3*r4*f1 */
/* (if you replace r3 by r2 both times in this move you get the same
 effect) */
		f(0);
		rCCW(2); rCCW(3); f(0); f(7);
		rCW(0); rCCW(3); f(7);
		rCW(3); rCW(3); rCW(3); rCW(3); f(7);
		rCCW(0); rCW(3); f(7);
		rCW(3); rCW(3); rCW(3); rCW(3); f(0);
		rCW(2); rCW(3); f(0);
	}

	void polar2Swap34() {
/* polar2swap18=f1 *r3^(-1)*r4^(-1)*f3*f4 *r1*r4^(-1)*f4 *r4^4*f4
  *r1^(-1)*r4*f4 *r4^4*f3 *r3*r4*f1 */
		f(0);
		rCCW(2); rCCW(3); f(6); f(5);
		rCW(0); rCCW(3); f(5);
		rCW(3); rCW(3); rCW(3); rCW(3); f(5);
		rCCW(0); rCW(3); f(5);
		rCW(3); rCW(3); rCW(3); rCW(3); f(6);
		rCW(2); rCW(3); f(0);
	}

	void equator2Swap16() {
/* equator2swap36=f1 *r4^(-1)*r3^(-1)*f1*f2 *r2*r3^(-1)*f2 *r3^4*f2
  *r2^(-1)*r3*f2 *r3^4*f1 *r4*r3*f1 */
		f(0);
		rCCW(3); rCCW(2); f(0); f(7);
		rCW(1); rCCW(2); f(7);
		rCW(2); rCW(2); rCW(2); rCW(2); f(7);
		rCCW(1); rCW(2); f(7);
		rCW(2); rCW(2); rCW(2); rCW(2); f(0);
		rCW(3); rCW(2); f(0);
	}

	void equator2Swap34() {
/* equator2swap18=f1 *r4^(-1)*r3^(-1)*f3*f4 *r2*r3^(-1)*f4 *r3^4*f4
 *r2^(-1)*r3*f4 *r3^4*f3 *r4*r3*f1 */
		f(0);
		rCCW(3); rCCW(2); f(6); f(5);
		rCW(1); rCCW(2); f(5);
		rCW(2); rCW(2); rCW(2); rCW(2); f(5);
		rCCW(1); rCW(2); f(5);
		rCW(2); rCW(2); rCW(2); rCW(2); f(6);
		rCW(3); rCW(2); f(0);
	}

/* 2 * 98 moves -> 4 out, correct PDF? */
	void nsMidSwap() {
/* NSmid_swap=r3*r2*f1 *r2*f1 *r3*f1 *r2^(-1)*f1 *r3^(-2)*f1
  *r2*f1 *r3*f1 *r2^(-1)*f1 *r3^4*f1
  *r2*f1 *r3*f1 *r2^(-1)*f1 *r3^(-2)*f1 *r2*f1
  *r3^2*f1 *r2^(-1)*f1 *r3^2*f1 *r2*f1
  *r3^(-1)*f1 *r2^(-1)*f1 *r3^(-1)*f1 *r2*f1
  *r3*f1 *r2^(-1)*f1 *r2*r3^(-2)*f1 *r2^(-1)*f1
  *r3^(-1)*f1 *r2*f1 *r2^(-1)*f1 *r3*f1 *r2*f1
  *r2^(-1)*f1 *r2*f1 *r3^2*f1
  *r2^(-1)*f1 *r2*f1 *r2^(-1)*f1 *r2*r3^(-1)*f1
  *r2^(-1)*f1 *r3*f1 *r2*f1 *r2^(-2)*r3^2 */
		rCW(2); rCW(1); f(0);
		rCW(1); f(0);
		rCW(2); f(0);
		rCCW(1); f(0);
		rCCW(2); rCCW(2); f(0);
		rCW(1); f(0);
		rCW(2); f(0);
		rCCW(1); f(0);
		rCW(2); rCW(2); rCW(2); rCW(2); f(0);
		rCW(1); f(0);
		rCW(2); f(0);
		rCCW(1); f(0);
		rCCW(2); rCCW(2); f(0);
		rCW(1); f(0);
		rCW(2); rCW(2); f(0);
		rCCW(1); f(0);
		rCW(2); rCW(2); f(0);
		rCW(1); f(0);
		rCCW(2); f(0);
		rCCW(1); f(0);
		rCCW(2); f(0);
		rCW(1); f(0);
		rCW(2); f(0);
		rCCW(1); f(0);
		rCW(1); rCCW(2); rCCW(2); f(0);
		rCCW(1); f(0);
		rCCW(2); f(0);
		rCW(1); f(0);
		rCCW(1); f(0);
		rCW(2); f(0);
		rCW(1); f(0);
		rCCW(1); f(0);
		rCW(1); f(0);
		rCW(2); rCW(2); f(0);
		rCCW(1); f(0);
		rCW(1); f(0);
		rCCW(1); f(0);
		rCW(1); rCCW(2); f(0);
		rCCW(1); f(0);
		rCW(2); f(0);
		rCW(1); f(0);
		rCCW(1); rCCW(1); rCW(2); rCW(2);
	}

/* 2 * 98 moves -> 4 out, correct PDF? */
	void nsSwap() {
/* NS_swap=r4*r1*f1 *r1*f1 *r4*f1 *r1^(-1)*f1 *r4^(-2)*f1
  *r1*f1 *r4*f1 *r1^(-1)*f1 *r4^44*f1
  *r1*f1 *r4*f1 *r1^(-1)*f1 *r4^(-2)*f1 *r1*f1
  *r4^2*f1 *r1^(-1)*f1 *r4^2*f1 *r1*f1
  *r4^(-1)*f1 *r1^(-1)*f1 *r4^(-1)*f1 *r1*f1
  *r4*f1 *r1^(-1)*f1 *r1*r4^(-2)*f1 *r1^(-1)*f1
  *r4^(-1)*f1 *r1*f1 *r1^(-1)*f1 *r4*f1 *r1*f1
  *r1^(-1)*f1 *r1*f1 *r4^2*f1
  *r1^(-1)*f1 *r1*f1 *r1^(-1)*f1 *r1*r4^(-1)*f1
  *r1^(-1)*f1 *r4*f1 *r1*f1 *r1^(-2)*r4^2 */
		rCW(3); rCW(0); f(0);
		rCW(0); f(0);
		rCW(3); f(0);
		rCCW(0); f(0);
		rCCW(3); rCCW(3); f(0);
		rCW(0); f(0);
		rCW(3); f(0);
		rCCW(0); f(0);
		rCW(3); rCW(3); rCW(3); rCW(3); f(0);
		rCW(0); f(0);
		rCW(3); f(0);
		rCCW(0); f(0);
		rCCW(3); rCCW(3); f(0);
		rCW(0); f(0);
		rCW(3); rCW(3); f(0);
		rCCW(0); f(0);
		rCW(3); rCW(3); f(0);
		rCW(0); f(0);
		rCCW(3); f(0);
		rCCW(0); f(0);
		rCCW(3); f(0);
		rCW(0); f(0);
		rCW(3); f(0);
		rCCW(0); f(0);
		rCW(0); rCCW(3); rCCW(3); f(0);
		rCCW(0); f(0);
		rCCW(3); f(0);
		rCW(0); f(0);
		rCCW(0); f(0);
		rCW(3); f(0);
		rCW(0); f(0);
		rCCW(0); f(0);
		rCW(0); f(0);
		rCW(3); rCW(3); f(0);
		rCCW(0); f(0);
		rCW(0); f(0);
		rCCW(0); f(0);
		rCW(0); rCCW(3); f(0);
		rCCW(0); f(0);
		rCW(3); f(0);
		rCW(0); f(0);
		rCCW(0); rCCW(0); rCW(3); rCW(3);
	}

/* 2 * 152 moves -> 3 out, correct? */
	void ewSwap() {
/* EW_swap=f2*f3*f5 *r2*f5 *r1*r2*r3*r4*f5
  *r2*f5 *r3^(-1)*f5 *r2^(-1)*f5 *r3^(-1)*f5
  *r2*f5 *r3^(-1)*f5 *r2^(-1)*f5 *r3^3*f5
  *r2*f5 *r3^(-1)*f5 *r2^(-1)*f5 *r3^2*f5 *r2*f5
  *r3^(-1)*f5 *r2^(-1)*f5 *r3^2*r2*f5
  *r2^(-1)*f5 *r2*f4 *r4^(-1)*r1^(-1)*f5 *r3^(-1)*f5 *r2^(-1)*f5
  *r3^(-1)*f5 *r2*f5 *r3*f5 *r2^(-1)*f5 *r3^2*f5
  *r2*f5 *r3^(-1)*f5 *r2^(-1)*f5 *r3^(-1)*f5 *r2*f5
  *r3*f5 *r2^(-1)*f5 *r2*f5 *r2^(-1)*f5
  *r3*r2^2*f5 *r2*f5 *r3^2*f5 *r2^(-1)*f5
  *r3^(-3)*f5 *r2*f5 *r3^2*f5
  *r2^(-1)*f5 *r3^(-1)*f5 *r2^(-1)*f5 *r2*f5 *r2^(-1)*f5
  *r2*r3^(-1)*f5 *r2^(-1)*f5 *r3^(-1)*f5 *r2*f5
  *r3*f5 *r2^(-1)*f5 *r3*f5 *r2*f5 *r2^(-1)*f5
  *r2*f5 *r2^(-2)*r3*f5 *r2*f5
  *r2^(-2)*f5 *r2^(-1)*f5*f3*f2 */
		f(7); f(6); f(4);
		rCW(1); f(4);
		rCW(0); rCW(1); rCW(2); rCW(3); f(4);
		rCW(1); f(4);
		rCCW(2); f(4);
		rCCW(1); f(4);
		rCCW(2); f(4);
		rCW(1); f(4);
		rCCW(2); f(4);
		rCCW(1); f(4);
		rCW(2); rCW(2); rCW(2); f(4);
		rCW(1); f(4);
		rCCW(2); f(4);
		rCCW(1); f(4);
		rCW(2); rCW(2); f(4);
		rCW(1); f(4);
		rCCW(2); f(4);
		rCCW(1); f(4);
		rCW(2); rCW(2); f(4);
		rCW(1); f(4);
		rCCW(1); f(4);
		rCW(1); f(5);
		rCCW(3); rCCW(0); f(4);
		rCCW(2); f(4);
		rCCW(1); f(4);
		rCCW(2); f(4);
		rCW(1); f(4);
		rCW(2); f(4);
		rCCW(1); f(4);
		rCW(2); rCW(2); f(4);
		rCW(1); f(4);
		rCCW(2); f(4);
		rCCW(1); f(4);
		rCCW(2); f(4);
		rCW(1); f(4);
		rCW(2); f(4);
		rCCW(1); f(4);
		rCW(1); f(4);
		rCCW(1); f(4);
		rCW(2); rCW(1); rCW(1); f(4);
		rCW(1); f(4);
		rCW(2); rCW(2); f(4);
		rCCW(1); f(4);
		rCCW(2); rCCW(2); rCCW(2); f(4);
		rCW(1); f(4);
		rCW(2); rCW(2); f(4);
		rCCW(1); f(4);
		rCCW(2); f(4);
		rCCW(1); f(4);
		rCW(1); f(4);
		rCCW(1); f(4);
		rCW(1); rCCW(2); f(4);
		rCCW(1); f(4);
		rCCW(2); f(4);
		rCW(1); f(4);
		rCW(2); f(4);
		rCCW(1); f(4);
		rCW(2); f(4);
		rCW(1); f(4);
		rCCW(1); f(4);
		rCW(1); f(4);
		rCCW(1); rCCW(1); rCW(2); f(4);
		rCW(1); f(4);
		rCCW(1); rCCW(1); f(4);
		rCCW(1); f(4); f(6); f(7);
	}

/* 5 * 107 moves, 5 out (30 * 107 moves, restored)
  I think this was meant to be from rainbo12.log but rainbo13.log copied here. */
	void badLong3Cycle() {
/*badLong3cycle=r4^(-2)*f3 *r4*f4 *r4*f4 *r4^(-1)*r1^2*f3
  *r4^(-1)*f3 *r1^(-3)*f3 *r4^(-1)*f3
  *r1*f3 *r4^2*f3 *r1*f3 *r4^(-1)*f3
  *r1^(-1)*f3 *r4^2*f3 *r1*f3
  *r4^(-1)*f3 *r1^(-1)*f3 *r4^4*f3
  *r1*f3 *r4*f3 *r1^(-1)*f3 *r4*f3
  *r1^(-2)*f3 *r4^(-1)*f3 *r1*f3 *r4^(-1)*f3 *r1^(-1)*f3
  *r4^2*f3 *r1^4*f3 *r4*f3
  *r1^(-3)*f3 *r4^(-1)*f3 *r1*f3 *r4^2*f3
  *r1^(-1)*f3 *r4*f3 *r1^3*f3 *r4*f3
  *r1^(-3)*f3 *r4^(-1)*f3
  *r1^2*f3 *r4^(-1)*f3 *r1^(-1)*f3
*/
		rCCW(3); rCCW(3); f(6);
		rCW(3); f(5);
		rCW(3); f(5);
		rCCW(3); rCW(0); rCW(0); f(6);
		rCCW(3); f(6);
		rCCW(0); rCCW(0); rCCW(0); f(6);
		rCCW(3); f(6);
		rCW(0); f(6);
		rCW(3); rCW(3); f(6);
		rCW(0); f(6);
		rCCW(3); f(6);
		rCCW(0); f(6);
		rCW(3); rCW(3); f(6);
		rCW(0); f(6);
		rCCW(3); f(6);
		rCCW(0); f(6);
		rCW(3); rCW(3); rCW(3); rCW(3); f(6);
		rCW(0); f(6);
		rCW(3); f(6);
		rCCW(0); f(6);
		rCW(3); f(6);
		rCCW(0); rCCW(0); f(6);
		rCCW(3); f(6);
		rCW(0); f(6);
		rCCW(3); f(6);
		rCCW(0); f(6);
		rCW(3); rCW(3); f(6);
		rCW(0); rCW(0); rCW(0); rCW(0); f(6);
		rCW(3); f(6);
		rCCW(0); rCCW(0); rCCW(0); f(6);
		rCCW(3); f(6);
		rCW(0); f(6);
		rCW(3); rCW(3); f(6);
		rCCW(0); f(6);
		rCW(3); f(6);
		rCW(0); rCW(0); rCW(0); f(6);
		rCW(3); f(6);
		rCCW(0); rCCW(0); rCCW(0); f(6);
		rCCW(3); f(6);
		rCW(0); rCW(0); f(6);
		rCCW(3); f(6);
		rCCW(0); f(6);
	}

/* 379 moves, small 3 & 4 wedges swapped */
	void long2Cycle() {
/*long2cycle=f1*f3 *r1^(-1)*f3*f1 *r4^(-2)*f1*f3 *r1*f3*f1
  *r4^(-1)*f1*f3 *r1^(-1)*f3*f1 *r4^(-1)*f1*f3 *r1^(-1)*f3*f1
  *r4^(-1)*f1*f3 *r1^2*f3*f1 *r4^(-1)*f1*f3 *r1^(-1)*f3*f1
  *r4*f1*f3 *r1*f3*f1 *r4^(-3)*f1*f3 *r1^(-1)*f3*f1 *r4^(-1)*f1*f3
  *r1^(-1)*f3*f1 *r4^(-1)*f1*f3 *r1^2*f3*f1 *r4^(-3)*f1*f3
  *r1^(-1)*f3*f1 *r4^(-1)*f1*f3 *r1^(-1)*f3*f1
  *r4^(-1)*f1*f3 *r1*f3*f1 *r4^(-1)*f1*f3 *r1*f3*f1 *r4*f1*f3
  *r1^(-2)*f3*f1 *r4*f1*f3 *r1*f3*f1 *r4*f1*f3
  *r1*f3*f1 *r4*f1*f3 *r1^(-1)*f3*f1 *r4*f1*f3
  *r1*f3*f1 *r4*f1*f3 *r1^(-2)*f3*f1 *r4*f1*f3
  *r1*f3*f1 *r4*f1*f3 *r1*f3*f1 *r4*f1*f3
  *r1^(-1)*f3*f1 *r4^2*f1*f3 *r1*f3*f1
  *r4^(-3)*f1*f3 *r1^(-1)*f3*f1 *r4^(-1)*f1*f3
  *r1^(-1)* f3*f1 *r4^(-1)*f1*f3 *r1^2*f3*f1 *r4^(-3)*f1*f3
  *r1^(-1)*f3*f1 *r4^(-1)*f1*f3 *r1^(-1)*f3*f1
  *r4^(-1)*f1*f3 *r1*f3*f1 *r4^(-1)*f1*f3 *r1*f3*f1
  *r4*f1*f3 *r1^(-2)*f3*f1 *r4*f1*f3
  *r1*f3*f1 *r4*f1*f3 *r1*f3*f1 *r4*f1*f3
  *r1^(-1)*f3*f1 *r4^2*f1*f3 *r1^(-1)*f3*f1
  *r4*f1*f3 *r1*f3*f1 *r4*f1*f3 *r1*f3*f1
  *r4^3*f1*f3 *r1^(-2)*f3*f1 *r4*f1*f3 *r1*f3*f1
  *r4*f1*f3 *r1*f3*f1 *r4^3*f1*f3 *r1^(-1)*f3*f1
  *r4^(-1)*f1*f3 *r1*f3*f1 *r4*f1*f3
  *r1^(-2)*f3*f1 *r4*f1*f3 *r1*f3*f1 *r4*f1*f3
  *r1*f3*f1 *r4*f1*f3 *r1^(-1)*f3*f1
  *r4^2*f1*f3 *r1*f3*f1 *r4^(-2)*f1*f3 *r1^(-1)*f3*f1
  *r4*f1*f3 *r1*f3*f1 *r4^(-2)*f1*f3
  *r1^(-1)*f3*f1 *r4^(-1)*f1*f3 *r1^(-1)*f3*f1
  *r4^(-1)*f1*f3 *r1^2*f3*f1 *r4^(-3)*f1*f3 *r1^(-1)*f3*f1
  *r4^(-1)*f1*f3 *r1^(-1)*f3*f1 *r4^(-1)*f1*f3
  *r1^2*f3*f1 *r4^(-1)*f1*f3 *r1^(-1)*f3*f1
  *r4^(-1)*f1*f3 *r1*f3*f1 *r4^2
*/
		f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); rCCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); rCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCCW(3); rCCW(3); rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); rCW(0); f(5); f(7);
		rCCW(3); rCCW(3); rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCW(3); rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCCW(3); rCCW(3); rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); rCW(0); f(5); f(7);
		rCCW(3); rCCW(3); rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCW(3); rCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); rCW(3); rCW(3); f(7); f(5);
		rCCW(0); rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); rCW(3); rCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCW(3); rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCCW(3); rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCCW(3); rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); rCW(0); f(5); f(7);
		rCCW(3); rCCW(3); rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); rCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCCW(0); f(5); f(7);
		rCCW(3); f(7); f(5);
		rCW(0); f(5); f(7);
		rCW(3); rCW(3);
	}

	void solveFirstWedge() {
		/* Find yellow */
		int pos = findColor(0, 0);
		int tz = wedgeFromPosition(pos);
		int b = bandFromPosition(pos);
		int i, lat;

		if (b > 0) {
			fPuzzle(0);
			tz = (3 * mball.wedges / 2 - tz - 1) % mball.wedges;
		}
		if (tz < mball.wedges / 2) {
			for (i = 0; i < tz; i++) {
				rPuzzleCCW(0);
			}
		} else {
			for (i = 0; i < mball.wedges - tz; i++) {
				rPuzzleCW(0);
			}
		}
		for (lat = 1; lat < (mball.bands + 1) / 2; lat++) {
			pos = findColor(0, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			if (b >= mball.bands / 2) {
				if (tz == 0) {
					rCW(b);
					tz++;
				}
				if (tz < mball.wedges / 2) {
					f(1);
					tz = mball.wedges / 2 + 1 - tz;
				} else {
					f(mball.wedges / 2);
					tz = mball.wedges - 1 - tz + mball.wedges / 2;
				}
				//b = mball.bands - 1 - b;
			}
			if (tz < mball.wedges / 2) {
				for (i = 0; i < tz; i++) {
					rCCW(lat);
				}
			} else {
				for (i = 0; i < mball.wedges - tz; i++) {
					rCW(lat);
				}
			}
		}
		for (lat = mball.bands / 2; lat < mball.bands; lat++) {
			pos = findColor(0, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			if (b < mball.bands / 2) {
				if (tz < mball.wedges / 2) {
					f(1);
					tz = mball.wedges / 2 + 1 - tz;
				} else {
					f(mball.wedges / 2);
					tz = mball.wedges - 1 - tz + mball.wedges / 2;
				}
				//b = mball.bands - 1 - b;
			}
			if (tz < mball.wedges / 2) {
				for (i = 0; i < tz; i++) {
					rCCW(lat);
				}
			} else {
				for (i = 0; i < mball.wedges - tz; i++) {
					rCW(lat);
				}
			}
		}
	}

	void solveSecondWedge(int lon) {
		int pos = findColor(lon, 0);
		int tz = wedgeFromPosition(pos);
		int b = bandFromPosition(pos);
		int lat;

		while (tz > 4) {
			f(tz - 3);
			tz = tz - 3;
			b = mball.bands - 1 - b;
		}
		if ((tz == 1 || tz == 2) && b >= mball.bands / 2) {
			f(1);
			if (tz == 2)
				tz = tz + 1;
			else /* tz == 1 */
				tz = tz + 3;
			b = mball.bands - 1 - b;
		}
		if ((tz == 3 || tz == 4) && b >= mball.bands / 2) {
			f(1);
			if (tz == 4) {
				tz = tz - 3;
			} else { /* tz == 3 */
				tz = tz - 1;
			}
			//b = mball.bands - 1 - b;
		}
		/* b < mball.bands / 2 */
		if (tz == 3 || tz == 4) {
			f(2);
			f(1);
			tz = tz - 2;
		}
		if (tz == 2) {
			rCCW(0);
			f(1);
			rCW(0);
			f(1);
			//tz = 1;
		}
		for (lat = 1; lat < (mball.bands + 1) / 2; lat++) {
			pos = findColor(lon, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			while (tz > 4) {
				f(tz - 3);
				tz = tz - 3;
				b = mball.bands - 1 - b;
			}
			if ((tz == 1 || tz == 2) && b >= mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(1);
				rNPoleCW(lat - 1);
				if (tz == 1)
					tz = 4;
				else
					tz = 3;
				b = mball.bands - 1 - b;
			}
			if ((tz == 3 || tz == 4) && b >= mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(1);
				rNPoleCW(lat - 1);
				if (tz == 4)
					tz = 1;
				else
					tz = 2;
				//b = mball.bands - 1 - b;
			}
			/* b < mball.bands / 2 */
			if (tz == 2 || tz == 3 || tz == 4) {
				/*rnCCW(1, tz - 1);
				f(1);
				rnCW(1, tz - 1);
				f(1);*/
				f(tz);
				rnCW(lat, tz - 1);
				f(tz);
				rnCCW(lat, tz - 1);
				//tz = 1;
			}
		}
		for (lat = (mball.bands + 1) / 2; lat < mball.bands; lat++) {
			pos = findColor(lon, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			while (tz > 4) {
				f(tz - 3);
				tz = tz - 3;
				b = mball.bands - 1 - b;
			}
			if (tz == 2 && b < mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(1);
				rNPoleCW(lat - 1);
				tz = 4;
				b = mball.bands - 1 - b;
			}
			if ((tz == 3 || tz == 4) && b < mball.bands / 2) {
				rCW(lat);
				f(2);
				rCCW(lat);
				//b = mball.bands - 1 - b;
				if (tz == 4)
					tz = 2;
			}
			/* b >= mball.bands / 2 */
			if (tz == 2 || tz == 3 || tz == 4) {
				f(tz);
				rnCW(lat, tz - 1);
				f(tz);
				rnCCW(lat, tz - 1);
				//tz = 1;
			}
		}
	}

	void solveThirdWedge(int lon) {
		int pos = findColor(lon, 0);
		int tz = wedgeFromPosition(pos);
		int b = bandFromPosition(pos);
		int lat;

		while (tz > 5) {
			f(tz - 3);
			tz = tz - 3;
			b = mball.bands - 1 - b;
		}
		if ((tz == 2 || tz == 3) && b >= mball.bands / 2) {
			f(2);
			if (tz == 3)
				tz = tz + 1;
			else /* tz == 2 */
				tz = tz + 3;
			b = mball.bands - 1 - b;
		}
		if ((tz == 4 || tz == 5) && b >= mball.bands / 2) {
			f(2);
			if (tz == 5) {
				tz = tz - 3;
			} else { /* tz == 4 */
				tz = tz - 1;
			}
			//b = mball.bands - 1 - b;
		}
		/* b < mball.bands / 2 */
		if (tz == 4 || tz == 5) {
			f(3);
			f(2);
			tz = tz - 2;
		}
		if (tz == 3) {
			rCCW(0);
			f(2);
			rCW(0);
			f(2);
			//tz = 1;
		}
		for (lat = 1; lat < (mball.bands + 1) / 2; lat++) {
			pos = findColor(lon, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			while (tz > 5) {
				f(tz - 3);
				tz = tz - 3;
				b = mball.bands - 1 - b;
			}
			if ((tz == 2 || tz == 3) && b >= mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(2);
				rNPoleCW(lat - 1);
				if (tz == 2)
					tz = 5;
				else
					tz = 4;
				b = mball.bands - 1 - b;
			}
			if ((tz == 4 || tz == 5) && b >= mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(2);
				rNPoleCW(lat - 1);
				if (tz == 5)
					tz = 2;
				else
					tz = 3;
				//b = mball.bands - 1 - b;
			}
			/* b < mball.bands / 2 */
			if (tz == 3 || tz == 4) {
				/*rnCCW(lat, tz - 1);
				f(1);
				rnCW(lat, tz - 1);
				f(1);*/
				f(tz);
				rnCW(lat, tz - 2);
				f(tz);
				rnCCW(lat, tz - 2);
				//tz = 1;
			} else if (tz == 5) {
				f(4);
				rnCW(lat, 2);
				f(4);
				rnCCW(lat, 2);
				f(3);
				rCW(lat);
				f(3);
				rCCW(lat);
				//tz = 1;
			}
		}
		for (lat = (mball.bands + 1) / 2; lat < mball.bands; lat++) {
			pos = findColor(lon, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			while (tz > 5) {
				f(tz - 3);
				tz = tz - 3;
				b = mball.bands - 1 - b;
			}
			if (tz == 3 && b < mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(2);
				rNPoleCW(lat - 1);
				tz = 5;
				b = mball.bands - 1 - b;
			}
			if ((tz == 4 || tz == 5) && b < mball.bands / 2) {
				rCW(lat);
				f(3);
				rCCW(lat);
				//b = mball.bands - 1 - b;
				if (tz == 5)
					tz = 3;
			}
			/* b >= mball.bands / 2 */
			if (tz == 3 || tz == 4) {
				f(tz);
				rnCW(lat, tz - 2);
				f(tz);
				rnCCW(lat, tz - 2);
				//z = 1;
			} else if (tz == 5) {
				f(4);
				rnCW(lat, 2);
				f(4);
				rnCCW(lat, 2);
				f(3);
				rCW(lat);
				f(3);
				rCCW(lat);
				//tz = 1;
			}
		}
	}

	void solveForthWedge(int lon) {
		int pos = findColor(lon, 0);
		int tz = wedgeFromPosition(pos);
		int b = bandFromPosition(pos);
		int i, lat;

		while (tz > 6) {
			f(tz - 3);
			tz = tz - 3;
			b = mball.bands - 1 - b;
		}
		if ((tz == 3 || tz == 4) && b >= mball.bands / 2) {
			f(3);
			if (tz == 4)
				tz = tz + 1;
			else /* tz == 3 */
				tz = tz + 3;
			b = mball.bands - 1 - b;
		}
		if ((tz == 5 || tz == 6) && b >= mball.bands / 2) {
			f(3);
			if (tz == 6) {
				tz = tz - 3;
			} else { /* tz == 5 */
				tz = tz - 1;
			}
			//b = mball.bands - 1 - b;
		}
		/* b < mball.bands / 2 */
		if ((tz == 5 || tz == 6)) {
			f(4);
			f(3);
			tz = tz - 2;
		}
		if (tz == 4) {
			rCCW(0);
			f(3);
			rCW(0);
			f(3);
			//tz = 1;
		}
		for (lat = 1; lat < (mball.bands + 1) / 2; lat++) {
			pos = findColor(lon, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			while (tz > 6) {
				f(tz - 3);
				tz = tz - 3;
				b = mball.bands - 1 - b;
			}
			if ((tz == 3 || tz == 4) && b >= mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(3);
				rNPoleCW(lat - 1);
				if (tz == 3)
					tz = 6;
				else
					tz = 5;
				b = mball.bands - 1 - b;
			}
			if ((tz == 5 || tz == 6) && b >= mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(3);
				rNPoleCW(lat - 1);
				if (tz == 6)
					tz = 3;
				else
					tz = 4;
				//b = mball.bands - 1 - b;
			}
			/* b < mball.bands / 2 */
			if (tz >= 4 && tz <= 6) {
				for (i = 0; i < tz - 3; i++) {
					f(4);
					rCW(lat);
					f(4);
					rCCW(lat);
				}
				//tz = 1;
			}
		}
		for (lat = (mball.bands + 1) / 2; lat < mball.bands; lat++) {
			pos = findColor(lon, lat);
			tz = wedgeFromPosition(pos);
			b = bandFromPosition(pos);
			while (tz > 6) {
				f(tz - 3);
				tz = tz - 3;
				b = mball.bands - 1 - b;
			}
			if (tz == 4 && b < mball.bands / 2) {
				rNPoleCCW(lat - 1);
				f(3);
				rNPoleCW(lat - 1);
				tz = 6;
				b = mball.bands - 1 - b;
			}
			if ((tz == 5 || tz == 6) && b < mball.bands / 2) {
				rCW(lat);
				f(4);
				rCCW(lat);
				//b = mball.bands - 1 - b;
				if (tz == 6)
					tz = 4;
			}
			/* b >= mball.bands / 2 */
			if (tz >= 4 && tz <= 6) {
				for (i = 0; i < tz - 3; i++) {
					f(4);
					rCW(lat);
					f(4);
					rCCW(lat);
				}
				//tz = 1;
			}
		}
	}

	void solveWedgeOpp(int wedge, int band) {
		int b = mball.bands - 1 - band;

		rCCW(b);
		f(wedge);
		rCW(b);
	}

	boolean solveOppSameLat(int band) {
		int wedge = findOppSameLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveOppSameLat: wedge " +
					wedge + ", band " + band);
			solveWedgeOpp(wedge, band);
			return true;
		}
		return false;
	}

	boolean solveAdjOppLat(int band) {
		int wedge = findAdjOppLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveAdjOppLat: wedge " +
					wedge + ", band " + band);
			fo(wedge, mball.wedges / 2 + 1);
			return (solveOppSameLat(band) ||
				solveOppSameLat(mball.bands - 1 - band));
		}
		return false;
	}

	boolean solveNextAdjSameLat(int band) {
		int wedge = findNextAdjSameLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveNextAdjSameLat: wedge " +
					wedge + ", band " + band);
			switch (mball.wedges) {
			case 6:
				flipFor6_4s((wedge + 3) % mball.wedges, band);
				break;
			case 8:
				fo(wedge, 1);
				fo(wedge, 2);
				break;
			default:
				System.out.println("solveNextAdjSameLat: need to work out for wedge " +
					wedge);
				break;
			}
			return (solveOppSameLat(band) ||
				solveOppSameLat(mball.bands - 1 - band));
		}
		return false;
	}

	boolean solveAdjSameLat(int band) {
		int wedge = findAdjSameLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveAdj: wedge " +
					wedge + ", band " + band);
			switch (mball.wedges) {
			case 4:
				flipFor4d((wedge + 1) % mball.wedges,
					mball.bands - 1 - band);
				return true;
			case 6:
				fo(wedge, 1);
				fo(wedge, 2);
				return (solveOppSameLat(band) ||
					solveOppSameLat(mball.bands - 1 - band));
			case 8:
				fo(wedge, 1);
				rCW(mball.bands - 1 - band);
				fo(wedge, 2);
				rCCW(mball.bands - 1 - band);
				return (solveNextAdjSameLat(band));
			default:
				System.out.println("solveAdjSameLat: need to work out for wedge " +
					wedge);
				break;
			}
		}
		return false;
	}

	boolean solveAdjOppOppLat(int band) {
		int wedge = findAdjOppOppLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveAdjOppOppLat: wedge " +
					wedge + ", band " + band);
			switch (mball.wedges) {
			case 4:
				fo(wedge, 2);
				return (solveAdjSameLat(band));
			case 6:
				fo(wedge, 2);
				return (solveOppSameLat(band) ||
					solveOppSameLat(mball.bands - 1 - band));
			case 8:
				fo(wedge, 2);
				return (solveOppSameLat(band));
			default:
				System.out.println("solveAdjOppOppLat: need to work out for wedge " +
					wedge);
				break;
			}
		}
		return false;
	}

	boolean solveOppOppLat(int band) {
		int wedge = findOppOppLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveOppOppLat: wedge " +
					wedge + ", band " + band);
			f(wedge);
			return (solveAdjSameLat(mball.bands - 1 - band));
		}
		return false;
	}

	boolean solveAdjOppSameLat(int band) {
		int wedge = findAdjOppSameLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveAdjOppSameLat: wedge " +
					wedge + ", band " + band);
			fo(wedge, 2);
			return (solveOppOppLat(band));
		}
		return false;
	}

	boolean solveNextAdjOppLat(int band) {
		int wedge = findNextAdjOppLat(band);

		if (wedge != mball.wedges) {
			if (DEBUG)
				System.out.println("solveNextAdjOppLat: wedge " +
					wedge + ", band " + band);
			switch (mball.wedges) {
			case 6:
				fo(wedge, 1);
				return (solveNextAdjSameLat(band));
			case 8:
				fo(wedge, 2);
				return (solveAdjOppSameLat(band));
			default:
				System.out.println("solveNextAdjOppLat: need to work out for wedge " +
					wedge);
				break;
			}
		}
		return false;
	}

	void easyFlip(int refTz, int refB, int slice) {
		int pos = findColor(slice, refB);
		int tz = wedgeFromPosition(pos);

		if (DEBUG) {
			System.out.println("easyFlip for " + mball.wedges +
				": tz " + refTz);
		}
		switch (mball.wedges) {
		case 4:
			if (ringAdd(tz, -refTz) == 2 + slice) {
				fo(tz, -1);
				tz = ringAdd(tz, -1);
				if (DEBUG) {
					System.out.println("  moving 2 out " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 1 + slice) {
				fo(tz, -1);
				tz = ringAdd(tz, -1);
				if (DEBUG) {
					System.out.println("  moving 2 in " + tz);
				}
			}
			break;
		case 6:
			{
			int d = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
			if (ringAdd(tz, -refTz) == 4 + slice) {
				fo(tz, -2);
				tz = ringAdd(tz, -2);
				d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  moving 2 out " + tz);
				}
			}
			/* easier to take care of direction now, probably for all cases
			where (mball.wedges / 2) % 2 == 1 */
			if (ringAdd(tz, -refTz) == slice && d == 1) {
				f(tz);
				tz = ringAdd(tz, 2);
				d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  take 2 so to rotate " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 2 + slice && d == 0) {
				fo(tz, -1);
				//d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  rotating 2 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 2 + slice) {
				fo(tz, -2);
				tz = ringAdd(tz, -2);
				//d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  moving 2 in " + tz);
				}
			}
			break;
			}
		case 8:
			if (ringAdd(tz, -refTz) > 4 + slice) {
				fo(tz, -3);
				tz = ringAdd(tz, -3);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving >4 " + tz);
				}
			}
			if (slice >= 3) {
				if (ringAdd(tz, -refTz) == 4 + slice) {
					fo(tz, -3);
					tz = ringAdd(tz, -3);
					if (DEBUG) {
						System.out.println("  " + slice + " e: moving 4 " + tz);
					}
				}
			} else {
				if (ringAdd(tz, -refTz) == 4 + slice) {
					fo(tz, -2);
					tz = ringAdd(tz, -1);
					if (DEBUG) {
						System.out.println("  " + slice + ": moving 4 " + tz);
					}
				}
			}
			if (ringAdd(tz, -refTz) == 1 + slice) {
				fo(tz, -1);
				tz = ringAdd(tz, 1);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving 1 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 2 + slice) {
				fo(tz, -1);
				tz = ringAdd(tz, 1);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving 2 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 3 + slice) {
				fo(tz, -3);
				tz = ringAdd(tz, -3);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving 3 " + tz);
				}
			}
			break;
		case 10:
			{
			int d = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
			if (ringAdd(tz, -refTz) == 8 + slice) {
				fo(tz, -4);
				tz = ringAdd(tz, -4);
				d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  moving 2 out " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 6 + slice &&
					slice >= 3) {
				fo(tz, -4);
				tz = ringAdd(tz, -4);
				d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  moving 4 out " + tz);
				}
			} else if (ringAdd(tz, -refTz) == 6 + slice) {
				fo(tz, -3);
				tz = ringAdd(tz, -2);
				d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  moving 2 out small step " + tz);
				}
			}
			/* easier to take care of direction now, probably for all cases
			where (mball.wedges / 2) % 2 == 1 */
			if (ringAdd(tz, -refTz) == slice && d == 1) {
				f(tz);
				tz = ringAdd(tz, 4);
				d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  take 2 out so to rotate " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 2 + slice) {
				fo(tz, -1);
				tz = ringAdd(tz, 2);
				d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  backing up 2 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 4 + slice &&
				d == 0 && slice >= 4) {
				fo(tz, -3);
				fo(tz, -4);
				fo(tz, -3);
				//d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  rotating 5 " + tz);
				}
			} else if (ringAdd(tz, -refTz) == 4 + slice && d == 0) {
				fo(tz, -2);
				//d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  rotating 2 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 4 + slice) {
				fo(tz, -4);
				tz = ringAdd(tz, -4);
				//d = (d == 0) ? 1 : 0;
				if (DEBUG) {
					System.out.println("  moving 2 in " + tz);
				}
			}
			break;
			}
		case 12:
			if (ringAdd(tz, -refTz) == 8 && slice < 3) {
				fo(tz, -4);
				tz = ringAdd(tz, -3);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving 8 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) >= 6 + slice) {
				fo(tz, -5);
				tz = ringAdd(tz, -5);
				if (DEBUG) {
					System.out.println("  " + slice + " e: moving 8 " + tz);
				}
			}
			if (slice >= 5) {
				if (ringAdd(tz, -refTz) == 6 + slice) {
					fo(tz, -5);
					tz = ringAdd(tz, -5);
					if (DEBUG) {
						System.out.println("  " + slice + " e: moving 4 " + tz);
					}
				}
			} else {
				if (ringAdd(tz, -refTz) == 6 + slice) {
					fo(tz, -3);
					tz = ringAdd(tz, -1);
					if (DEBUG) {
						System.out.println("  " + slice + ": moving 4 " + tz);
					}
				}
			}
			if (ringAdd(tz, -refTz) == 1 + slice) {
				fo(tz, -1);
				tz = ringAdd(tz, 3);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving 1 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 2 + slice) {
				fo(tz, -1);
				tz = ringAdd(tz, 3);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving 2 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 3 + slice) {
				fo(tz, -2);
				tz = ringAdd(tz, 1);
				if (DEBUG) {
					System.out.println("  " + slice + ": moving 3 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 4 + slice && slice == 5) {
				fo(tz, -3);
				fo(tz, -4);
				fo(tz, -3);
				fo(tz, -4);
				tz = ringAdd(tz, -4);
				if (DEBUG) {
					System.out.println("  moving 5 " + tz);
				}
			} else if (ringAdd(tz, -refTz) == 4 + slice) {
				fo(tz, -2);
				tz = ringAdd(tz, 1);
				if (DEBUG) {
					System.out.println("  moving 4 " + tz);
				}
			}
			if (ringAdd(tz, -refTz) == 5 + slice) {
				fo(tz, -5);
				tz = ringAdd(tz, -5);
				if (DEBUG) {
					System.out.println("  moving 5a " + tz);
				}
			}
			break;
		default:
			break;
		}
	}

	void orientWithinWedge() {
		int wedge, band;

		/* user for bands > 1 */
		if (mball.wedges == 4) {
			for (band = 0; band < mball.bands / 2; band++) {
				for (wedge = mball.wedges - 1; wedge >= 0; wedge--) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction == 1) {
						flipFor4w(wedge, band);
					}
				}
			}
		} if (mball.wedges == 6) {
			for (band = 0; band < mball.bands / 2; band++) {
				for (wedge = mball.wedges - 1; wedge >= 0; wedge--) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction == 1) {
						flipFor6w((wedge + mball.wedges - 1) % mball.wedges, band);
					}
				}
			}
			/* if (allTop(w)) */
			for (wedge = 0; wedge < mball.wedges; wedge++) {
				int tz;

				for (tz = 0; tz < mball.wedges; tz++) {
					for (band = 0; band < mball.bands / 2; band++) {
						int pos, pos1, tz1;

						tz1 = (tz + mball.wedges - 1) % mball.wedges;
						pos1 = mball.wedgeLoc[tz1 + mball.wedges * band].wedge;
						pos = mball.wedgeLoc[tz + mball.wedges * band].wedge;
						if ((pos1 > pos) && (pos != 0)) {
							flipFor6_4(tz, band);
						}
					}
				}
			}
		} else if (mball.wedges == 8) {
			/* flipping opposite wedges i.e. 4 wedges apart */
			for (band = 0; band < mball.bands / 2; band++) {
				for (wedge = 0; wedge < mball.wedges / 2; wedge++) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction == 1 &&
							mball.wedgeLoc[ringAdd(wedge, mball.wedges / 2) + mball.wedges * band].direction == 1) {
						flipOppSegments(wedge, band);
					}
				}
			}
			/* flipping wedges 3 wedges apart */
			for (band = 0; band < mball.bands / 2; band++) {
				for (wedge = mball.wedges - 1; wedge >= 0; wedge--) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction == 1 &&
							mball.wedgeLoc[ringAdd(wedge, 3) + mball.wedges * band].direction == 1) {
						fo(wedge, 3);
						flipOppSegments(wedge, band);
						fo(wedge, 3);
					}
				}
			}
			/* flipping wedges 1 wedge apart */
			for (band = 0; band < mball.bands / 2; band++) {
				for (wedge = mball.wedges - 1; wedge >= 0; wedge--) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction == 1 &&
							mball.wedgeLoc[ringAdd(wedge, 1) + mball.wedges * band].direction == 1) {
						fo(wedge, 1);
						flipOppSegments(wedge, band);
						fo(wedge, 1);
					}
				}
			}
			/* flipping wedges 2 wedges apart, done later as less efficient */
			for (band = 0; band < mball.bands / 2; band++) {
				for (wedge = mball.wedges - 1; wedge >= 0; wedge--) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction == 1 &&
							mball.wedgeLoc[ringAdd(wedge, 2) + mball.wedges * band].direction == 1) {
						fo(wedge, 1);
						fo(wedge, 2);
						flipOppSegments(wedge, band);
						fo(wedge, 2);
						fo(wedge, 1);
					}
				}
			}
			/* lone wedges that need flipping, least efficient, so do last, 35 moves */
			for (band = 0; band < mball.bands / 2; band++) {
				for (wedge = mball.wedges - 1; wedge >= 0; wedge--) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction == 1) {
						flipWedgeSegment1(wedge, band);
					}
				}
			}
		}
	}

/* 3 approaches to solving */
	void solveLon() {
		/* solve wedges approach, half implemented */
		solveFirstWedge();
		solveSecondWedge(1);
		solveThirdWedge(2);
		solveForthWedge(3);
	}

	void solvePoles() {
		/* solve top to bottom approcach, not implemented */
	}

	void solveLat() {
		/* solve top and bottom segments */
		int iter, wedge, band, lat;

		if (mball.wedges == 2) {
			int notWedge, midBand;

			wedge = mball.generator.nextInt(2); /* better to pick most correct one */
			notWedge = (wedge == 0) ? 1 : 0;
			lat = (mball.bands - 1) / 2;
			for (band = 0; band < mball.bands; band++) {
				if (band != lat &&
						(mball.wedgeLoc[wedge + mball.wedges * lat].wedge !=
						mball.wedgeLoc[wedge + mball.wedges * band].wedge) &&
						(mball.wedgeLoc[wedge + mball.wedges * lat].wedge ==
						mball.wedgeLoc[notWedge + mball.wedges * band].wedge)) {
					rCW(band);
				}
			}
			if (!mball.checkSolved()) {
				f(wedge);
			}
			for (band = 0; band < mball.bands; band++) {
				if (band != lat &&
						(mball.wedgeLoc[wedge + mball.wedges * lat].wedge !=
						mball.wedgeLoc[wedge + mball.wedges * band].wedge) &&
						(mball.wedgeLoc[wedge + mball.wedges * lat].wedge ==
						mball.wedgeLoc[notWedge + mball.wedges * band].wedge)) {
					rCW(band);
				}
			}
			if (mball.orient && !mball.checkSolved()) {
				/*swap half, f, fix, f, swap bottom half */
				rNHalfHalf();
				f(wedge);
				for (band = 0; band < mball.bands; band++) {
					if (mball.wedgeLoc[wedge + mball.wedges * band].direction != 1) {
						rCW(band);
					}
				}
				f(wedge);
				rSHalfHalf();
				midBand = mball.bands / 2;
				if (mball.bands % 2 != 0) {
					if (mball.wedgeLoc[wedge].wedge !=
							mball.wedgeLoc[wedge + mball.wedges * midBand].wedge) {
						rCW(midBand);
					}
					if (mball.wedgeLoc[wedge + mball.wedges * midBand].direction !=
							mball.wedgeLoc[wedge].direction) {
						rCW(midBand);
						f(wedge);
						rSHalfHalf();
						f(wedge);
						rCW(midBand);
						rSHalfHalf();
						f(wedge);
						rSHalfHalf();
					}
				}
			}
			return;
		}
		/* Stage 1 and Stage 2, align same sized segments */
		for (band = 0; band < mball.bands / 2; band++) {
			int ob = mball.bands - 1 - band;
			boolean fixing = !checkBandSolved(band);

			while (fixing) {
				if ((solveOppSameLat(band)) || (solveOppSameLat(ob)))
					continue;
				/* How to Make Segments Opposite */
				/* Option 1 */
				if ((solveAdjOppLat(band)) || (solveAdjOppLat(ob)))
					continue;
				/* Option 2 */
				if (mball.wedges == 4 || mball.wedges == 6) {
					if (solveAdjSameLat(band))
						continue;
					if (solveAdjSameLat(ob))
						continue;
				}
				if (mball.wedges == 6 && solveOppOppLat(band)) {
					continue;
				}
				if (mball.wedges == 8) {
					if ((solveNextAdjSameLat(band)) || (solveNextAdjSameLat(ob)))
						continue;
				}
				if (mball.wedges == 6 || mball.wedges == 8) {
					if ((solveAdjOppOppLat(band)) || (solveAdjOppOppLat(ob)))
						continue;
				}
				if (mball.wedges == 6) {
					if ((solveNextAdjSameLat(band)) || (solveNextAdjSameLat(ob)))
						continue;
				}
				if (mball.wedges == 6) {
					if ((solveNextAdjOppLat(band)) || (solveNextAdjOppLat(ob)))
						continue;
				}
				/* Option 3 */
				if (mball.wedges == 8) {
					if ((solveAdjSameLat(band)) || (solveAdjSameLat(ob)))
						continue;
				}
				if (mball.wedges != 6 && solveOppOppLat(band)) {
					continue;
				}
				if (mball.wedges == 8) {
					if ((solveAdjOppSameLat(band)) || (solveAdjOppSameLat(ob))
							|| (solveNextAdjOppLat(band)) || (solveNextAdjOppLat(ob)))
						continue;
				}
				fixing = false;
			}
		}
		/* Stage 3, align the different sized segments to center band,
		   do it this way as there is an odd number of mball.bands.  WSD site
		   moves the larger segments instead, not much difference. */
		for (iter = 0; iter < 2 + ((mball.wedges < 8) ? 2 : 0); iter++) {
			/* need to go through again as unsolved wedges may be moving around */
			for (wedge = 0; wedge < mball.wedges; wedge++) {
				lat = (mball.bands - 1) / 2;
				for (band = lat - 1; band >= 0; band--) {
					int color, pos, tz, diff;
					boolean bypass = false;

					if (checkSegmentSolved(wedge, band, lat)) {
						if (DEBUG)
							System.out.println("solved wedge " +
								wedge + ", band " + band);
						continue;
					}
					color = mball.wedgeLoc[wedge + mball.wedges * lat].wedge;
					pos = findColor(color, band);
					tz = wedgeFromPosition(pos);
					diff = (mball.wedges - wedge + tz + mball.wedges / 2) % mball.wedges;
					if (DEBUG) {
						System.out.println("wedge " +
							wedge + ", band " + band);
						System.out.println("tz " +
							tz + ", diff " + diff);
					}
					/*if (diff != 0 && wedge == 0) {
						if (diff < mball.wedges / 2) {
							rnCCW(band, diff);
							rnCCW(mball.bands - 1 - band, diff);
						} else {
							diff = mball.wedges - diff;
							rnCW(band, diff);
							rnCW(mball.bands - 1 - band, diff);
						}
					}*/
					if (diff == 1) {
						switch (mball.wedges) {
						case 6:
							bypass = true;
							flipFor6_4sm((wedge + mball.wedges - 1) % mball.wedges, band);
							break;
						case 4:
							fo(wedge, mball.wedges - 2);
							break;
						default:
							fo(wedge, mball.wedges / 2 - 1);
							break;
						}
					} else if (diff == mball.wedges - 1) {
						if (mball.wedges == 6) {
							flipFor6_4sm((wedge + 1) % mball.wedges, band);
							bypass = true;
						} else
							fo(wedge, 2);
					} else if (diff == 2) {
						switch (mball.wedges) {
						case 8:
							fo(wedge, mball.wedges / 2);
							fo(wedge, mball.wedges / 2 - 1);
							break;
						case 6:
							fo(wedge, mball.wedges / 2);
							break;
						default:
							fo(wedge, mball.wedges - 1);
							fo(wedge, mball.wedges - 2);
							break;
						}
					} else if (diff == mball.wedges - 2) {
						fo(wedge, 1);
						if (mball.wedges != 6)
							fo(wedge, 2);
					} else if (diff == 3) {
						fo(wedge, mball.wedges / 2);
					} else if (diff == mball.wedges - 3) {
						fo(wedge, 1);
					}
					if (!bypass) {
						rCCW(band);
						f((wedge + mball.wedges / 2) % mball.wedges);
						rCW(band);
						rCCW(mball.bands - 1 - band);
						f((wedge + mball.wedges / 2) % mball.wedges);
						rCW(mball.bands - 1 - band);
					}
				}
			}
		}
		if (mball.orient) {
			int refPos = findColor(0, (mball.bands + 1) / 2 - 1);
			int refTz = wedgeFromPosition(refPos);
			int refB = bandFromPosition(refPos);
			int pos, tz, d;

			if (DEBUG) {
				System.out.println("Orient: wedge " + refTz + ", band " + refB);
			}
			/* put numbers on top, no ordering needed for 2 in a ring,
			 * otherwise need logic because ordering on opposite side
			 * is reversed */
			if (mball.wedges > 2 &&
					mball.wedgeLoc[refTz + mball.wedges * refB].direction == 1) {
				fPuzzle(0); /* should not have but seems to double the work otherwise */
				if (refTz >= mball.wedges / 2)
					refTz = mball.wedges - 1 - refTz + mball.wedges / 2;
				else
					refTz = mball.wedges / 2 - 1 - refTz;
				refB = mball.bands - 1 - refB;
				if (DEBUG) {
					System.out.println("righted, wedge " + refTz + ", band " + refB);
				}
			}
			/* order wedges, the first half is "easy" */
			for (tz = 1; tz < mball.wedges / 2; tz++)
				easyFlip(refTz, refB, tz);
			if (DEBUG) {
				System.out.println(mball.wedges + " wedge specific");
			}
			switch (mball.wedges) {
			case 2:
				if (mball.wedgeLoc[1].direction != mball.wedgeLoc[0].direction) {
				f(mball.wedgeLoc[1].wedge);
				}
				break;
			case 4:
				pos = findColor(mball.wedges / 2, refB);
				tz = wedgeFromPosition(pos);
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
				fo(tz, -1);
				}
				break;
			case 6:
				pos = findColor(mball.wedges / 2, refB);
				tz = wedgeFromPosition(pos);
				d = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
					fo(tz, -2);
					tz = ringAdd(tz, -2);
					d = (d == 0) ? 1 : 0;
				}
				if (d == 1) {
					fo(tz, -2);
					f(tz);
					fo(tz, -1);
					f(tz);
					fo(tz, -2);
					fo(tz, -1);
				}
				break;
			case 8:
				int tz4, tz7;
				/* check if we can squeeze out another easy flip */
				pos = findColor(mball.wedges / 2, refB);
				tz4 = wedgeFromPosition(pos);
				pos = findColor(mball.wedges - 1, refB);
				tz7 = wedgeFromPosition(pos);
				if (ringAdd(tz4, -refTz) == mball.wedges - 1) {
					fo(tz4, -3);
					tz = ringAdd(tz4, -tz7);
					tz4 = ringAdd(tz4, -3);
					tz7 = (tz4 + tz) % mball.wedges;
				} else if (ringAdd(tz7, -refTz) == mball.wedges / 2) {
					f(tz7);
					tz = ringAdd(tz7, -tz4);
					tz7 = ringAdd(tz7, 3);
					tz4 = ringAdd(tz7, tz);
				} else if ((ringAdd(tz7, -refTz) == mball.wedges - 2) &&
						(ringAdd(tz4, -refTz) != mball.wedges / 2)) {
					fo(tz7, -2);
					tz = ringAdd(tz7, -tz4);
					tz7 = ringAdd(tz7, -1);
					tz4 = ringAdd(tz7, tz);
				}
				/* check if its the 2 pair pattern */
				if ((ringAdd(tz7, -refTz) == mball.wedges / 2 + 1) &&
					(ringAdd(tz4, -refTz) != mball.wedges / 2)) {
				saturn2Pair(ringAdd(tz7, -1));
				}
				/* fix for 3 out */
				if (ringAdd(tz4, -refTz) == mball.wedges / 2) {
					tz = ringAdd(tz7, -refTz);
					if (tz == mball.wedges / 2 + 1) {
						saturn3Jump(ringAdd(tz4, 1));
					} else if (tz == mball.wedges - 2) {
						saturn3JumpBack(ringAdd(tz4, 1));
					}
				} else if (ringAdd(tz7, -refTz) == mball.wedges - 1) {
					tz = (tz4 - refTz + mball.wedges) % mball.wedges;
					if (tz == mball.wedges / 2 + 1) {
						saturn3Jump(ringAdd(tz7, -3));
					} else if (tz == mball.wedges - 2) {
						saturn3JumpBack(ringAdd(tz7, -3));
					}
				}
				/* if even, outermost 6 & 7 could be swapped. */
				if (mball.bands % 2 == 0) {
					pos = findColor(mball.wedges / 2 + 1, refB);
					tz = wedgeFromPosition(pos);
					if (ringAdd(tz, -refTz) != mball.wedges / 2 + 1) {
						swapEven(ringAdd(tz, -1));
					}
				}
				break;
			case 10:
				int d1, d2;
				/* I wrote a program to find these, see puck.c in java dist */
				pos = findColor(mball.wedges / 2, refB);
				tz = wedgeFromPosition(pos);
				d = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
					fo(tz, -4);
					tz = ringAdd(tz, -4);
					d = (d == 0) ? 1 : 0;
				} else if (ringAdd(tz, -refTz) == mball.wedges / 2 + 2) {
					fo(tz, -2);
					fo(tz, -6);
					fo(tz, -2);
					fo(tz, -6);
					fo(tz, -2);
					fo(tz, -6);
					f(tz);
					fo(tz, -6);
					f(tz);
					tz = ringAdd(tz, -2);
					d = (d == 0) ? 1 : 0;
				}
				if (d == 1) {
					fo(tz, -2);
					fo(tz, 1);
					fo(tz, -2);
					fo(tz, 1);
					fo(tz, -2);
					fo(tz, 1);
				}
				pos = findColor(mball.wedges / 2 + 1, refB);
				tz = wedgeFromPosition(pos);
				d = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
				if (ringAdd(tz, -refTz) == mball.wedges - 2) {
					fo(tz, -2);
					fo(tz, -6);
					fo(tz, -2);
					fo(tz, -6);
					fo(tz, -2);
					fo(tz, -6);
					f(tz);
					fo(tz, -6);
					f(tz);
					tz = ringAdd(tz, -2);
					d = (d == 0) ? 1 : 0;
				}
				if (d == 1) {
					fo(tz, -2);
					fo(tz, 1);
					fo(tz, -2);
					fo(tz, 1);
					fo(tz, -2);
					fo(tz, 1);
				}
				pos = findColor(mball.wedges / 2 + 2, refB);
				d = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
				pos = findColor( mball.wedges - 2, refB);
				d1 = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
				pos = findColor(mball.wedges - 1, refB);
				d2 = mball.wedgeLoc[(pos % mball.wedges) + ((mball.bands + 1) / 2 - 1) * mball.wedges].direction;
				if (d == 1 && d2 == 1) {
					fo(tz, 2);
					fo(tz, -4);
					fo(tz, 3);
					fo(tz, 1);
					fo(tz, -5);
					fo(tz, -4);
					fo(tz, -3);
					fo(tz, -1);
					fo(tz, 2);
					fo(tz, 3);
				} else if (d == 1 && d1 == 1) {
					fo(tz, -1);
					f(tz);
					fo(tz, 1);
					fo(tz, 2);
					f(tz);
					fo(tz, -1);
					f(tz);
					fo(tz, -1);
					fo(tz, 1);
					fo(tz, 2);
				} else if (d1 == 1 && d2 == 1) {
				f(tz);
				fo(tz, 1);
				fo(tz, 2);
				fo(tz, 3);
				fo(tz, 1);
				f(tz);
				fo(tz, 1);
				f(tz);
				fo(tz, 2);
				fo(tz, 3);
				}
				break;
			case 12:
				/* check if we can squeeze out another easy flip */
				/* TODO: take into account that C wedge may
				be solved */
				/*				int tz6, tzc;

				pos = findColor(mball.wedges / 2, refB);
				tz6 = wedgeFromPosition(pos);
				pos = findColor(mball.wedges - 1, refB);
				tzc = wedgeFromPosition(pos);
				if (ringAdd(tz6, -refTz) == mball.wedges - 1) {
					fo(tz6, -5);
					tz = ringAdd(tz6, -tzc);
					tz6 = ringAdd(tz6, -5);
					tzc = ringAdd(tz6, tz);
				} else if (ringAdd(tzc, -refTz) == mball.wedges / 2) {
					f(tzc);
					tz = ringAdd(tzc, -tz6);
					tzc = ringAdd(tzc, 5);
					tz6 = ringAdd(tzc, tz);
				}
				*/
				/* http://www.jaapsch.net/puzzles/puck.htm */
				/* What Jaap calls Segment 6 and down: etc. */
				/* I am solving in order, he solves in reverse. */
				/*
				6: 6: !
				5: ceacae	  -> 8 to 7 f(-2,-4,0,-2,0,-4)
				-> 8 to 7 f(-4,0,-2,0,-4,-2)
				4: dbcdcd (r?) -> 9 to 7 f(-2,-1,-2,-1,0,-2)
				3: adbcdcd	 -> A to 7 f(0,-2,-1,-2,-1,0,-2)
				2: dcdcbd (r?) -> B to 7 f(-2,0,-1,-2,-1,-2)
				1: a	   -> C to 7 f(0)
				*/
				pos = findColor(mball.wedges / 2, refB);
				tz = wedgeFromPosition(pos);
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
					fo(tz, -5);
				} else if (ringAdd(tz, -refTz) == mball.wedges - 2) {
					fo(tz, -2);
					fo(tz, -4);
					fo(tz, -3);
					fo(tz, -2);
					fo(tz, -3);
					fo(tz, -2);
				} else if (ringAdd(tz, -refTz) == mball.wedges - 3) {
					fo(tz, -3);
					fo(tz, -1);
					fo(tz, -2);
					fo(tz, -1);
					fo(tz, -2);
					fo(tz, -3);
					fo(tz, -1);
				} else if (ringAdd(tz, -refTz) == mball.wedges / 2 + 2) {
					f(tz);
					fo(tz, -1);
					f(tz);
					fo(tz, -1);
					fo(tz, -2);
					f(tz);
				} else if (ringAdd(tz, -refTz) == mball.wedges / 2 + 1) {
					if (mball.generator.nextInt(2) == 0) {
						fo(tz, 1);
						fo(tz, 3);
						fo(tz, -1);
						fo(tz, 1);
						fo(tz, -1);
						fo(tz, 3);
					} else {
						fo(tz, 3);
						fo(tz, -1);
						fo(tz, 1);
						fo(tz, -1);
						fo(tz, 3);
						fo(tz, 1);
					}
				}
				/*
				5: 5: !
				4: eadc deae -> 9 to 8 f(-1,1,-2,-1,0,-1,1,-1) r?
				3: cabcbc	-> A to 8 f(-2 -1 -2 -1,0,-2)
				2: cebc dcec -> B to 8 f(-3,1,-2,-1,-2,-3,1,-3)
				1: cbcbac	-> C to 8 f(-2,0,-1,-2,-1,-2)
				*/
				pos = findColor(mball.wedges / 2 + 1, refB);
				tz = wedgeFromPosition(pos);
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
					fo(tz, -2);
					fo(tz, -4);
					fo(tz, -3);
					fo(tz, -2);
					fo(tz, -3);
					fo(tz, -2);
				} else if (ringAdd(tz, -refTz) == mball.wedges - 2) {
					f(tz);
					fo(tz, -4);
					fo(tz, -1);
					fo(tz, -2);
					fo(tz, -1);
					f(tz);
					fo(tz, -4);
					f(tz);
				} else if (ringAdd(tz, -refTz) == mball.wedges - 3) {
					f(tz);
					fo(tz, -1);
					f(tz);
					fo(tz, -1);
					fo(tz, -2);
					f(tz);
				} else if (ringAdd(tz, -refTz) == mball.wedges / 2 + 2) {
					f(tz);
					fo(tz, 2);
					fo(tz, -1);
					f(tz);
					fo(tz, 1);
					f(tz);
					fo(tz, 2);
					f(tz);
				}
				/*
				4: 4: !
				3: dfcb abfd -> A to 9 f(-1,-3,0,-1,-2,-1,-3,-1)
				2: bdbc badb -> B to 9 f(-3,-5,-1,0,-1,-2,-5,-3)
				1: bdab cbdb -> C to 9 f(-3,-5,-2,-1,0,-1,-5,-3)
				-> C to 9 f(-1,-3,-1,-2,-1,0,-3,-1)
				*/
				pos = findColor(mball.wedges / 2 + 2, refB);
				tz = wedgeFromPosition(pos);
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
					if (mball.generator.nextInt(2) == 0) {
						f(tz);
						fo(tz, 2);
						fo(tz, -1);
						fo(tz, -2);
						fo(tz, -3);
						fo(tz, -2);
						fo(tz, 2);
						f(tz);
					} else {
					fo(tz, -2);
						f(tz);
						fo(tz, -2);
						fo(tz, -1);
						fo(tz, -2);
						fo(tz, -3);
						f(tz);
						fo(tz, -2);
					}
				} else if (ringAdd(tz, -refTz) == mball.wedges - 2) {
					fo(tz, 1);
					fo(tz, 3);
					fo(tz, -1);
					fo(tz, -2);
					fo(tz, -1);
					f(tz);
					fo(tz, 3);
					fo(tz, 1);
				} else if (ringAdd(tz, -refTz) == mball.wedges - 3) {
					f(tz);
					fo(tz, 2);
					fo(tz, -1);
					f(tz);
					fo(tz, 1);
					f(tz);
					fo(tz, 2);
					f(tz);
				}
				/*
				3: 3: !
				2: dbdab dbdab (r?)-> B to A f(-1,0,-3,-1,-3,-1,0,-3,-1,-3)
				1: badbd badbd (r?)-> C to A f(-3,-1,-3,0,-1,-3,-1,-3,0,-1)
				*/
				pos = findColor(mball.wedges - 3, refB);
				tz = wedgeFromPosition(pos);
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
					fo(tz, 1);
					fo(tz, -1);
					fo(tz, 1);
					fo(tz, -2);
					fo(tz, -1);
					fo(tz, 1);
					fo(tz, -1);
					fo(tz, 1);
					fo(tz, -2);
					fo(tz, -1);
				} else if (ringAdd(tz, -refTz) == mball.wedges - 2) {
					f(tz);
					fo(tz, -1);
					fo(tz, 2);
					f(tz);
					fo(tz, 2);
					f(tz);
					fo(tz, -1);
					fo(tz, 2);
					f(tz);
					fo(tz, 2);
				}
				/*
				2: 2: !
				1: abace deba eacd -> B to A f(0,-1,0,-2,-4,-3,-4,-1,0,-4,0-2,-3)
				*/
				pos = findColor(mball.wedges - 2, refB);
				tz = wedgeFromPosition(pos);
				if (ringAdd(tz, -refTz) == mball.wedges - 1) {
					if (mball.generator.nextInt(2) == 0) {
						fo(tz, 7);
						fo(tz, 6);
						fo(tz, 7);
						fo(tz, 5);
						fo(tz, 3);
						fo(tz, 4);
						fo(tz, 3);
						fo(tz, 6);
						fo(tz, 7);
						fo(tz, 3);
						fo(tz, 7);
						fo(tz, 5);
						fo(tz, 4);
					} else {
						fo(tz, 4);
						fo(tz, 5);
						fo(tz, 7);
						fo(tz, 3);
						fo(tz, 7);
						fo(tz, 6);
						fo(tz, 3);
						fo(tz, 4);
						fo(tz, 3);
						fo(tz, 5);
						fo(tz, 7);
						fo(tz, 6);
						fo(tz, 7);
					}
				}
				break;
			default:
				break;
			}
			/* Orient segments within wedges */
			orientWithinWedge();
		}
	}

	/* This procedure coordinates the solution process. */
	void solvePieces() {
		((MballFrame) frame).callback(
			MballInterface.ACTION_RESET);
/*
		quadrantizedPattern();
		bandedPattern();
		checkeredPattern();
		simpleSwap();
		nsMidSwap();
		nsSwap();
		ewSwap();
		swapAdj34();
		swapOp04(0, 0);
		polar2Swap16(); // non-oriented
		polar2Swap34();
		equator2Swap16();
		equator2Swap34();
		long2Cycle();
		long3Cycle();
		threeOut(1, 1);
		twoOut(0, 1);
		flipOppSegments(1, 1);
		flipWedgeSegment(2, 2);
		flipWedgeSegment1(0, 0);
		saturn2Pair(0);
		saturn3Jump(5);
		saturn3JumpBack(4);
		swapEven(0);
		swap4by4d();
		swap4by4();
		swap2by4();
		flipFor4w(0, 0);
		flipFor4d(0, 0);
		flipFor4a(0, 0);
		flipFor4_3(0, 0);
		swap4by4();
		swap4by4op(0, 0);
		flipFor6_4(0, 0);
		flipFor6_4sm(1, 0);
		flipFor6_4s(3, 0);
		flipFor6_3f(0, 0);
		flipFor6d(0, 0);
		flipFor6w(0, 0);
		flipFor6_3(0, 0);
		flipFor6_3r(0, 0);
*/
		if (!mball.checkSolved()) {
			solveLat();
		}
		((MballFrame) frame).callback(
			MballInterface.ACTION_COMPUTED);
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solvePieces();
					runFlag = false;
				}
			}
		}
	}

	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
